/*
Example code for the technique presented in the paper

GPU-based Sphere Tracing for Radial Basis Functions Implicits
(Batagelo, H. C.; Gois, J. P.). International Journal of Image and Graphics,
14(1-2), 2014, DOI 10.1142/S0219467814500041

Copyright (c) 2014, Harlen Costa Batagelo
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the copyright holder may not be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL HARLEN COSTA BATAGELO BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#version 420

in vec2 tex_coord;
layout(location = 0) out vec4 frgcolor;
layout(location = 1) out vec4 frgstat;

layout(binding = 0) uniform samplerBuffer tboSamplerP;
layout(binding = 1) uniform samplerBuffer tboSamplerC;
layout(binding = 2) uniform samplerCube texCubeMap;

struct SCamera
{
    vec3 position;
    float scale;
    float orthoA;
    float orthoB;
};

struct SRay
{
    vec3 origin;
    vec3 direction;
};

uniform int maxCentersPerCellPlusOne;
uniform int numberOfCells;
uniform vec4 pouDiv;
uniform float pouOverlap;
uniform vec3 minPou, maxPou;
uniform vec3 pouSz;
uniform vec3 pouCellSz;

uniform Centers
{
    vec4 lipCen[4096];
};

uniform SCamera camera;
uniform vec4 lightPosition;
uniform vec4 ambientProduct;
uniform vec4 diffuseProduct;
uniform vec4 specularProduct;
uniform float shininess;

uniform mat4 modelViewMatrix;
uniform mat4 modelViewMatrixInv;
uniform float minSphereDiameter;

const vec3 boxMinimum = vec3(-6.9731, -7.7, -4.2911);
const vec3 boxMaximum = vec3( 6.9731, 7.7, 9.6551);
const float sphereRadius = 15.0;

const vec3 axisX = vec3(1.0, 0.0, 0.0);
const vec3 axisY = vec3(0.0, 1.0, 0.0);
const vec3 axisZ = vec3(0.0, 0.0, 1.0);

const mat3 identity = mat3(1, 0, 0, 0, 1, 0, 0, 0, 1);

#define T 1.6180339887

vec3 hsvramp(in float t)
{
    vec3 rgb = vec3(0);
    float hdash = t * 6;
    float x = 1.0 - abs((mod(hdash, 2.0)) - 1.0);
    if (hdash < 1.0) rgb.rg = vec2(1.0, x);
    else if (hdash < 2.0) rgb.rg = vec2(x, 1.0);
    else if (hdash < 3.0) rgb.gb = vec2(1.0, x);
    else if (hdash < 4.0) rgb.gb = vec2(x, 1.0);
    else if (hdash < 5.0) rgb.rb = vec2(x, 1.0);
    else rgb.rb = vec2(1.0, x);

    return rgb;
}

float length2(in vec2 p)
{
    return sqrt(p.x*p.x + p.y*p.y);
}

float pows(in float x, in float y)
{
    float power = 0;
    if (x < 0)
        power = exp(y*log(-x));
    else
        power = exp(y*log(x));

    return power;
}

float lengthn(in vec2 p, in float n)
{
    return pows(pows(p.x,n) + pows(p.y,n), 1.0 / n);
}

float genericTorus(in vec3 p)
{
    vec2 t = vec2(3, 1.5);
    vec2 q = vec2(length2(p.xz) - t.x, p.y);
    return lengthn(q, 8) - t.y;
}

bool intersectPlane(in SRay ray, // Ray origin and direction
                    in vec3 n,   // Plane normal
                    in vec3 o,  // Point on the plane
                    out float t) // Time of intersection
{
    float vd = dot(n, ray.direction);
    if (vd < 0)
    {
        t = dot(n, o-ray.origin) / vd;
        if (t < 0)
            return false;
        return true;
    }
    return false;
}

/*
 * Intersects ray with box ( general case ).
 *
 * Function implements fastest algorithm for ray-box
 * intersection based on slab test. For details see:
 *
 * T. Kay, J. Kajiya. "Ray tracing complex scenes".
 * Computer Graphics, 20 (4), 269 - 278, 1986.
 */

bool intersectBox(in SRay ray     /* ray origin and direction */,
                  in vec3 minimum /* minimum point of a box */,
                  in vec3 maximum /* maximum point of a box */,
                  out float start /* time of 1st intersection */,
                  out float final /* time of 2nd intersection */)
{
    vec3 OMIN = (minimum - ray.origin) / ray.direction;
    vec3 OMAX = (maximum - ray.origin) / ray.direction;
    vec3 MAX = max(OMAX, OMIN);
    vec3 MIN = min(OMAX, OMIN);

    final = min(MAX.x, min(MAX.y, MAX.z));
    start = max(max(MIN.x, 0.0), max(MIN.y, MIN.z));

    return final > start;
}

/*
 * Intersects ray with sphere in coordinates origin.
 *
 * Function implements simple algorithm for ray-sphere
 * intersection based on equating of the equations of
 * ray and sphere. For details see:
 *
 * Ray Sphere Intersection ( CGS Wiki )
 * http://wiki.cgsociety.org/index.php/Ray_Sphere_Intersection
 */

bool intersectSphere(in SRay ray     /* ray origin and direction */,
                     in float radius /* sphere radius */,
                     out float start /* time of 1st intersection */,
                     out float final /* time of 2nd intersection */)
{
    float A = dot(ray.direction, ray.direction);
    float B = dot(ray.direction, ray.origin);
    float C = dot(ray.origin, ray.origin) - radius*radius;
    float D = B*B - A*C;
    start = final = 0;

    if (D > 0.0)
    {
        D = sqrt(D);
        start = max(0.0, (-B-D)/A);
        final = (-B+D)/A;
        return final > 0.0;
    }

    return false;
}

float sdBox(in vec3 p, in vec3 b)
{
    vec3 d = abs(p) - b;
    return min(max(d.x, max(d.y, d.z)), 0.0) +
            length(max(d, 0.0));
}

float sdPlane(in vec3 p, in vec3 n, in vec3 o)
{
    return dot(p-o, n);
}

float calcFunction(in vec3 p)
{    
    vec3 a = ((p - minPou) / pouSz) * pouDiv.xyz;
    vec3 ijk0 = floor(a);

    vec3 ovld = round(a) - a;
    vec3 ovls = sign(ovld * max(vec3(0), (vec3((pouOverlap-1)/2) - abs(ovld)) * ijk0 * (pouDiv.xyz-1 - ijk0)));

    vec4 ijk[8];
    ijk[0].w = 1.0;
    ijk[1].xw = ovls.xx;
    ijk[2].yw = ovls.yy;
    ijk[3].zw = ovls.zz;
    ijk[4].xzw = vec3(ovls.xz, ovls.x * ovls.z);
    ijk[5].yzw = vec3(ovls.yz, ovls.y * ovls.z);
    ijk[6].xyw = vec3(ovls.xy, ovls.x * ovls.y);
    ijk[7] = vec4(ovls, ovls.x * ovls.y * ovls.z);

    float f = 0, ac = 0;
    float lip = 0;    
    for (int nc = 0; nc < 8; ++nc)
    {
        vec4 cijk = ijk[nc];
        if (cijk.w == 0.0)
            continue;

        cijk.xyz += ijk0;
        int i = int(cijk.x * pouDiv.w + cijk.y * pouDiv.z + cijk.z);
        int k = i * maxCentersPerCellPlusOne;
        // x: Lipschitz, y: curNumCen, z: isValid
        vec2 lipAndCen = texelFetch(tboSamplerP, k).xy;
        vec3 s = cijk.xyz * pouCellSz + minPou;
        vec3 t = s + pouCellSz;
        vec3 ovl = ((pouOverlap-1)*(t-s))/2;

        s -= ovl;
        t += ovl;

        int l = k;
        float floc = 0;
        int m = int(lipAndCen.y);
        k++;
        for (int j = 0; j < m; ++j)
        {
            vec4 cC = texelFetch(tboSamplerC, l + j);

            vec3 df = p - texelFetch(tboSamplerP, k + j).xyz;
            float r = length(df);
            floc += (cC.x * r * r * r) - dot(cC.yzw, 3 * r * df);
        }
        vec4 poly = texelFetch(tboSamplerC, l + m);
        floc += poly.x + dot(poly.yzw, p);

        vec3 den = t-s;
        vec3 wei = (4*(p-s)*(t-p))/(den*den);
        float d = 1 - wei.x * wei.y * wei.z;
        float ddd = d * d * d;
        float dddd = ddd * d;
        float c = -6*dddd*d + 15*ddd*d - 10*ddd + 1;

        f += floc * c;
        lip += lipAndCen.x * c;
        ac += c;
    }

    f /= ac;

    return f / (lip / ac);
}

vec3 calcNormal(in vec3 p)
{
    vec3 a = ((p - minPou) / pouSz) * pouDiv.xyz;
    vec3 ijk0 = floor(a);

    vec3 ovld = round(a) - a;
    vec3 ovls = sign(ovld * max(vec3(0), (vec3((pouOverlap-1)/2) - abs(ovld)) * ijk0 * (pouDiv.xyz-1 - ijk0)));

    vec4 ijk[8];
    ijk[0].w = 1.0;
    ijk[1].xw = ovls.xx;
    ijk[2].yw = ovls.yy;
    ijk[3].zw = ovls.zz;
    ijk[4].xzw = vec3(ovls.xz, ovls.x * ovls.z);
    ijk[5].yzw = vec3(ovls.yz, ovls.y * ovls.z);
    ijk[6].xyw = vec3(ovls.xy, ovls.x * ovls.y);
    ijk[7] = vec4(ovls, ovls.x * ovls.y * ovls.z);

    vec3 g = vec3(0);
    for (int nc = 0; nc < 8; ++nc)
    {
        vec4 cijk = ijk[nc];
        if (cijk.w == 0.0)
            continue;

        cijk.xyz += ijk0;

        int i = int(cijk.x * pouDiv.w + cijk.y * pouDiv.z + cijk.z);
        int k = i * maxCentersPerCellPlusOne;
        vec3 s = cijk.xyz * pouCellSz + minPou;
        vec3 t = s + pouCellSz;
        vec3 ovl = ((pouOverlap-1)*(t-s))/2;
        s -= ovl;
        t += ovl;
        int l = i * maxCentersPerCellPlusOne;
        vec3 gloc = vec3(0);
        int curNumCen = int(texelFetch(tboSamplerP, k).y);
        k++;

        for (int j = 0; j < curNumCen; ++j)
        {
            vec4 cC = texelFetch(tboSamplerC, l + j);
            vec3 df = p - texelFetch(tboSamplerP, k + j).xyz;
            float r = length(df);
            gloc += (cC.x * 3 * r * df) - cC.yzw *
                    3 / r * (r * r * identity + outerProduct(df, df));
        }
        gloc += texelFetch(tboSamplerC, l + curNumCen).yzw;

        vec3 den = t-s;
        vec3 wei = (4*(p-s)*(t-p))/(den*den);
        float d = 1 - wei.x * wei.y * wei.z;
        float ddd = d * d * d;
        float dddd = ddd * d;
        float c = -6*dddd*d + 15*ddd*d - 10*ddd + 1;

        g += gloc * vec3(c);
    }

    return normalize(g);
}

float combCalcFunction(in vec3 p)
{
    if (p.x < minPou.x+0.001 || p.x > maxPou.x-0.001 ||
        p.y < minPou.y+0.001 || p.y > maxPou.y-0.001 ||
        p.z < minPou.z+0.001 || p.z > maxPou.z-0.001)
        return sdPlane(p, vec3(0,1,0), minPou);
    return calcFunction(p);
}

vec3 combCalcNormal(in vec3 p)
{
    if (p.x < minPou.x+0.001 || p.x > maxPou.x-0.001 ||
        p.y < minPou.y+0.001 || p.y > maxPou.y-0.001 ||
        p.z < minPou.z+0.001 || p.z > maxPou.z-0.001)
        return vec3(0,1,0);
    return calcNormal(p);
}

#define EPSILON 0.1 // Small value for moving origin of shadow ray
//#define SHADOWS // Use these macro for enable / disable shadows

/*
 * Intersects ray with implicit surface.
 */

bool intersectSurface(in SRay ray   /* ray origin and direction */,
                      in float ts   /* time when a ray enters a box */,
                      in float te   /* time when a ray leaves a box */,
                      out float th  /* time of ray-surface hit */,
                      out float it  /* number of iterations */)
{       
    th = -1;
    it = 0;

    float t = ts, fv;

    vec3 p = ray.origin + ray.direction * t;
    do
    {
        fv = combCalcFunction(p);// / Lipschitz;
        it++;

        float cutoff = (it*it*it)/1.25e8; // it^3/500^3
        if (fv < minSphereDiameter + cutoff)
        {
            th = t + fv + cutoff;
            return true;
        }

        t += fv;
        p += ray.direction * fv;
    } while (t < te);

    if (intersectPlane(ray, vec3(0,1,0), minPou, t))
    {
        th = t;
        return true;
    }
    return false;
}

/* Computes lighting in an intersection point and its final color */

vec3 phong(in vec3 P /* intersection point with surface */,
           in vec3 N /* normal to the surface in this point */,
           in float pt /* t value at P */)
{
    vec3 L = normalize(lightPosition.xyz - P);
    vec3 E = normalize(camera.position - P);
    vec3 R = normalize(2.0 * dot(L, N) * N - L);

    float NdotL = dot(N, L);
    float Kd = max(NdotL, 0.0);
    float Ks = (NdotL < 0.0) ? 0.0 : pow(max(dot(R, E), 0.0), shininess);

#if defined ( SHADOWS )
    SRay ray = SRay(P + L * EPSILON, L);
    float ts, te, t;
    if (intersectBox(ray, minPou+0.01, maxPou-0.01, ts, te))
    {
        float it;
        if (intersectSurface(ray, ts, te, t, it))
            Kd = Ks = 0;
    }
#endif

    // Compute ambient occlusion
    float ao = 0;
    for (float i = 1; i <= 3; ++i)
    {
        vec3 P2 = P + N * (i/15);
        if (P2.x < minPou.x+0.001 || P2.x > maxPou.x-0.001 ||
            P2.y < minPou.y+0.001 || P2.y > maxPou.y-0.001 ||
            P2.z < minPou.z+0.001 || P2.z > maxPou.z-0.001)
            break;

        float distn = length(P2-P);
        float dists = max(0, combCalcFunction(P2));
        ao += (1/pow(2,i))*max(0, distn-dists);
    }
    ao = 1-10*ao;
    ao = max(0,ao);

    float ft = max(0, pt-25);
    vec3 fogColor = vec3(1.0, 1.0, 1.0);
    float fog = 1/exp(ft*.025);

    vec3 texcube = vec3(1);
    // If outside POU, won't use cube mapping
    if (!(P.x < minPou.x+0.001 || P.x > maxPou.x-0.001 ||
          P.y < minPou.y+0.001 || P.y > maxPou.y-0.001 ||
          P.z < minPou.z+0.001 || P.z > maxPou.z-0.001))
    {
        L = vec3(0, 100, 0);
        vec3 R = normalize(2.0 * dot(L, N) * N - L);
        texcube = texcube * vec3(0.5) + texture(texCubeMap, -R).xyz * vec3(0.5);
    }
    else
    {
        return (ambientProduct.xyz + (vec3(0.6) * Kd * texcube) * ao) * fog + (1-fog) * fogColor;
    }

    return (ambientProduct.xyz +
            (diffuseProduct.xyz * Kd * texcube +
             specularProduct.xyz * Ks) * ao) * fog + (1-fog) * fogColor;
}

/*
 * Generates primary ray for a current fragment.
 *
 * You can change camera state by setting function argument.
 * It can be used for stereo rendering ( move camera left and
 * right for each fragment and tracing two primary rays ) or
 * anti-aliasing.
 */

SRay generateRay(in SCamera camera)
{
    vec2 coords = tex_coord.xy * camera.scale;
    vec3 direction = mat3x3(modelViewMatrixInv) * vec3(coords, -1);

    return SRay(camera.position, normalize(direction));
}

vec4 raytrace(in SCamera camera, out float depth)
{
    depth = 0;
    float ts, te;
    vec4 color = vec4(0);
    float t, it;

    SRay ray = generateRay(camera);

    bool intBox = false;
    if (intersectBox(ray, minPou+0.0011, maxPou-0.0011, ts, te))
        intBox = true;

    if (!intBox)
    {
        if (!intersectPlane(ray, vec3(0,1,0), minPou, t))
            discard;
    }
    else
        if (!intersectSurface(ray, ts, te, t, it))
            discard;

    if (t == -1)
        depth = -1;
    else
    {
        vec3 P = ray.origin + ray.direction * t;
        vec3 N = combCalcNormal(P);

        color = vec4(phong(P, N, t), 1.0);
        depth = (camera.orthoB * vec3(mat3x3(modelViewMatrix) * P).z -
                 camera.orthoA) * 0.5 + 0.5;
    }

    frgstat = vec4(it, 0, 0, 1);

    return color;
}

void main()
{
    float depth;
    frgcolor = raytrace(camera, depth);

    if (depth != -1)
        gl_FragDepth = depth;
}
