/*
Example code for the technique presented in the paper

GPU-based Sphere Tracing for Radial Basis Functions Implicits
(Batagelo, H. C.; Gois, J. P.). International Journal of Image and Graphics,
14(1-2), 2014, DOI 10.1142/S0219467814500041

Copyright (c) 2014, Harlen Costa Batagelo
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the copyright holder may not be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL HARLEN COSTA BATAGELO BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <QFile>
#include <QTextDocument>
#include <QFileDialog>
#include <QDir>
#include <QApplication>

#include "glwidget.h"
#include "util.h"

#ifdef WIN32
#undef max
#undef min
#endif

GLWidget::GLWidget(QWidget *parent) :
    QGLWidget(parent)
{
    tboP = 0;
    texTboP = 0;
    tboC = 0;
    texTboC = 0;

    ubo = 0;
    vboVertices = NULL;
    vboIndices = NULL;
    fbo = 0;

    shaderProgram = NULL;
    vertexShader = NULL;
    fragmentShader = NULL;
    currentShader = 0;

    fpsCounter = 0;
    mayRender = false;

    isFullScreen = false;
    showFPS = true;
    showModelViewMatrix = false;
    showShadows = false;
    showStatistics = false;

    numCenters = 0;
    centersP = NULL;
    centersN = NULL;
    centersC = NULL;

    tx_id[0] = -1; tx_id[1] = -1;tx_id[2] = -1;
    tx_id[3] = -1; tx_id[4] = -1; tx_id[5] = -1;
    tx_id[6] = -1;

    cube_img[0] = QImage(":/textures/negx.jpg","JPG");
    cube_img[1] = QImage(":/textures/negy.jpg","JPG");
    cube_img[2] = QImage(":/textures/negz.jpg","JPG");
    cube_img[3] = QImage(":/textures/posx.jpg","JPG");
    cube_img[4] = QImage(":/textures/posy.jpg","JPG");
    cube_img[5] = QImage(":/textures/posz.jpg","JPG");

    setAutoFillBackground(false);
}

GLWidget::~GLWidget()
{
    makeCurrent();

    destroyShadersAndBOs();
}

void GLWidget::initializeGL()
{
    if (!initializeOpenGLFunctions()) {
        QString str = QString("Sorry, cannot initialize OpenGL functions.\n"
                              "Make sure your graphics card is compatible with OpenGL 3.3.\n\n"
                              "The application will now exit.");
        showCriticalError(str);
        abort();
    }

    createOpenGLScene();

    qDebug("OpenGL version: %s\n", glGetString(GL_VERSION));
}

void GLWidget::resizeGL(int width, int height)
{    
    glViewport(0, 0, width, height);

    if (fbo) {
        glDeleteTextures(2, texo);
        glDeleteRenderbuffers(1, &rbo);
        glDeleteFramebuffers(1, &fbo);
        fbo = 0;
    }

    projectionMatrix = camera.ProjectionMatrix(width, height, -50.0, 1000.0);
    modelViewMatrix.setToIdentity();

    trackBall.resizeViewport(width, height);
    lightTrackBall.resizeViewport(width, height);

    // Compute pixel size in world space
    // This is the diameter of the minimum unbounding sphere
    minSphere = QVector2D(2.0 / static_cast<qreal>(width),
                          2.0 / static_cast<qreal>(height));

    createBkgndGradient();
}

void GLWidget::paintEvent(QPaintEvent *)
{
    QPainter painter(this);

    drawBkgnd(&painter);

    painter.beginNativePainting();
    drawOpenGLScene();
    painter.endNativePainting();

    displayInfo(&painter);

    painter.end();
}

void GLWidget::drawOpenGLScene()
{
    glPushClientAttrib(GL_CLIENT_PIXEL_STORE_BIT | GL_CLIENT_VERTEX_ARRAY_BIT);
    glPushAttrib(GL_ALL_ATTRIB_BITS);

    glEnable(GL_BLEND);
    glEnable(GL_DEPTH_TEST);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    if (mayRender) {
        if (!fbo) {
            // Create texture object
            glGenTextures(2, texo);
            glBindTexture(GL_TEXTURE_2D, texo[0]);
            glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F,
                         camera.viewportWidth(),
                         camera.viewportHeight(),
                         0, GL_RGBA, GL_FLOAT, 0);
            glBindTexture(GL_TEXTURE_2D, texo[1]);
            glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F,
                         camera.viewportWidth(),
                         camera.viewportHeight(),
                         0, GL_RGBA, GL_FLOAT, 0);
            glBindTexture(GL_TEXTURE_2D, 0);

            // Create render buffer object
            glGenRenderbuffers(1, &rbo);
            glBindRenderbuffer(GL_RENDERBUFFER, rbo);
            glRenderbufferStorage(GL_RENDERBUFFER,
                                  GL_DEPTH_COMPONENT,
                                  camera.viewportWidth(),
                                  camera.viewportHeight());

            // Create frame buffer object
            glGenFramebuffers(1, &fbo);
            glBindFramebuffer(GL_FRAMEBUFFER, fbo);

            glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT,
                                      GL_RENDERBUFFER, rbo);
            glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
                                   GL_TEXTURE_2D, texo[0], 0);
            glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT1,
                                   GL_TEXTURE_2D, texo[1], 0);

            glBindFramebuffer(GL_FRAMEBUFFER, 0);
        }

        glBindFramebuffer(GL_FRAMEBUFFER, fbo);

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        QMatrix4x4 rotationMatrix;
        rotationMatrix.setToIdentity();
        rotationMatrix.rotate(trackBall.getRotation());
        rotationMatrix = rotationMatrix.inverted();

        lastPosition = rotationMatrix * QVector3D(0.0, 0.0, 30.0);

        camera.SetLookAt(lastPosition,
                         QVector3D(0.0, 0.0, 0.0),
                         rotationMatrix * QVector3D(0.0, 1.0, 0.0));
        modelViewMatrix = camera.ModelViewMatrix();

        shaderProgram->bind();

        shaderProgram->setUniformValue("maxCentersPerCellPlusOne", maxCentersPerCell + 1);
        shaderProgram->setUniformValue("numberOfCells", numberOfCells);
        shaderProgram->setUniformValue("pouDiv", pouDiv);
        shaderProgram->setUniformValue("pouOverlap", pouOverlap);
        shaderProgram->setUniformValue("minPou", minPou);
        shaderProgram->setUniformValue("maxPou", maxPou);

        shaderProgram->setUniformValue("pouSz", pouSz);
        shaderProgram->setUniformValue("pouCellSz", pouCellSz);

        shaderProgram->setUniformValue("modelViewMatrix", modelViewMatrix);
        QMatrix4x4 modelViewMatrixInv = modelViewMatrix.inverted();
        shaderProgram->setUniformValue("modelViewMatrixInv", modelViewMatrixInv);
        shaderProgram->setUniformValue("projectionMatrix", projectionMatrix);
        shaderProgram->setUniformValue("minSphereDiameter",
                                       static_cast<GLfloat>(minSphere.y()/2.0));

        QVector4D ambientProduct = light.ambient * material.ambient;
        QVector4D diffuseProduct = light.diffuse * material.diffuse;
        QVector4D specularProduct = light.specular * material.specular;

        QMatrix4x4 lightRotationMatrix;
        lightRotationMatrix.setToIdentity();
        lightRotationMatrix.rotate(lightTrackBall.getRotation());

        QVector4D lightPosition = QVector4D(lightRotationMatrix *  QVector4D(0.0, 50.0, 20.0, 1.0));
        shaderProgram->setUniformValue("lightPosition", lightPosition);
        shaderProgram->setUniformValue("ambientProduct", ambientProduct);
        shaderProgram->setUniformValue("diffuseProduct", diffuseProduct);
        shaderProgram->setUniformValue("specularProduct", specularProduct);
        shaderProgram->setUniformValue("shininess",
                                       static_cast<GLfloat>(material.shininess));

        shaderProgram->setUniformValue("camera.position", camera.Position());
        shaderProgram->setUniformValue("camera.scale",
                                       static_cast<GLfloat>(camera.Scale()));
        shaderProgram->setUniformValue("camera.orthoA",
                                       static_cast<GLfloat>(camera.OrthoA()));
        shaderProgram->setUniformValue("camera.orthoB",
                                       static_cast<GLfloat>(camera.OrthoB()));

        glEnable(GL_TEXTURE_2D);
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_BUFFER, texTboP);
        glActiveTexture(GL_TEXTURE1);
        glBindTexture(GL_TEXTURE_BUFFER, texTboC);

        glActiveTexture(GL_TEXTURE2);

        for(int i = 0; i < 7; i++)
            if(tx_id[i] != -1) QGLWidget::deleteTexture(tx_id[i]);

        tx_id[6] = QGLWidget::bindTexture(cube_img[5],GL_TEXTURE_CUBE_MAP);
        glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
        glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, GL_TRUE);
        GLfloat maxAniso;
        glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &maxAniso);
        glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAX_ANISOTROPY_EXT, maxAniso);

        tx_id[0] = QGLWidget::bindTexture(cube_img[0],GL_TEXTURE_CUBE_MAP_NEGATIVE_X,GL_RGB);
        tx_id[1] = QGLWidget::bindTexture(cube_img[4],GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,GL_RGB);
        tx_id[2] = QGLWidget::bindTexture(cube_img[2],GL_TEXTURE_CUBE_MAP_NEGATIVE_Z,GL_RGB);
        tx_id[3] = QGLWidget::bindTexture(cube_img[3],GL_TEXTURE_CUBE_MAP_POSITIVE_X,GL_RGB);
        tx_id[4] = QGLWidget::bindTexture(cube_img[1],GL_TEXTURE_CUBE_MAP_POSITIVE_Y,GL_RGB);
        tx_id[5] = QGLWidget::bindTexture(cube_img[5],GL_TEXTURE_CUBE_MAP_POSITIVE_Z,GL_RGB);

        // Bind the buffer object to the uniform block
        if (ubo)
            glBindBuffer(GL_UNIFORM_BUFFER, ubo);

        vboVertices->bind();
        shaderProgram->enableAttributeArray("vPosition");
        shaderProgram->setAttributeBuffer("vPosition", GL_FLOAT, 0, 4, 0);

        vboIndices->bind();

        GLenum buffers[] = {GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1};
        glDrawBuffers(2, buffers);

        glDrawElements(GL_QUADS, 4, GL_UNSIGNED_INT, 0);

        vboIndices->release();
        vboVertices->release();
        glBindBuffer(GL_UNIFORM_BUFFER, 0);

        shaderProgram->release();

        if (showStatistics) {
            glReadBuffer(GL_COLOR_ATTACHMENT1);
            int bufferSize = camera.viewportWidth() * camera.viewportHeight();
            float *data = new float[bufferSize * 4];
            glReadPixels(0, 0, camera.viewportWidth(), camera.viewportHeight(),
                         GL_RGBA, GL_FLOAT, data);
            drawnPixels = 0;
            sumOfSteps = 0;
            maximumStep = 0;

            for (int i = 0; i < bufferSize * 4; i += 4) {
                float curDataIt = data[i];
                if (curDataIt) {
                    sumOfSteps += curDataIt;
                    drawnPixels++;
                    maximumStep = qMax(static_cast<int>(curDataIt), maximumStep);
                }
            }
            delete []data;
        }

        glBindFramebuffer(GL_FRAMEBUFFER, 0);

        // Draw FBO quad (deprecated code because I'm too lazy)
        glDisable(GL_DEPTH_TEST);
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, texo[0]);

        glBegin(GL_QUADS);
        glTexCoord2f(0, 0); glVertex2f(-1, -1);
        glTexCoord2f(0, 1); glVertex2f(-1,  1);
        glTexCoord2f(1, 1); glVertex2f( 1,  1);
        glTexCoord2f(1, 0); glVertex2f( 1, -1);
        glEnd();

        // Draw offscreen buffer
/*        if (showStatistics) {
            float offx = 20.0 / static_cast<float>(camera.viewportWidth()),
                  offy = 20.0 / static_cast<float>(camera.viewportHeight());
            glDisable(GL_BLEND);
            glBindTexture(GL_TEXTURE_2D, texo[1]);
            glBegin(GL_QUADS);
            glTexCoord2f(0, 0); glVertex2f(-1.0 + offx, -1.0 + offy);
            glTexCoord2f(0, 1); glVertex2f(-1.0 + offx, -0.5 + offy);
            glTexCoord2f(1, 1); glVertex2f(-0.5 + offx, -0.5 + offy);
            glTexCoord2f(1, 0); glVertex2f(-0.5 + offx, -1.0 + offy);
            glEnd();

            glBindTexture(GL_TEXTURE_2D, 0);
            glColor3f(1.0, 1.0, 1.0);
            glBegin(GL_LINE_LOOP);
            glVertex2f(-1.0 + offx, -1.0 + offy);
            glVertex2f(-1.0 + offx, -0.5 + offy);
            glVertex2f(-0.5 + offx, -0.5 + offy);
            glVertex2f(-0.5 + offx, -1.0 + offy);
            glEnd();
        } */
        glDisable(GL_TEXTURE_2D);
    }

    glPopAttrib();
    glPopClientAttrib();
}

void GLWidget::displayInfo(QPainter *painter)
{
    QRect rect;
    QString text;
    QTextDocument textDocument;

    if (showFPS) {
        fpsCounter++;
        if (fpsTime.elapsed() > 1e3) {
            fps = fpsCounter * 1e3 / fpsTime.elapsed();
            fpsTime.restart();
            fpsCounter = 0;
        }

        text.sprintf("FPS: %.1f", fps);

        QFont font = textDocument.defaultFont();
        font.setPointSize(width()/60);
        font.setWeight(QFont::DemiBold);
        painter->setFont(font);
        QFontMetrics metrics = QFontMetrics(font);
        int border = qMax(6, metrics.leading());

        rect = metrics.boundingRect(0, 0, width(), height(),
                                    Qt::AlignCenter |
                                    Qt::TextWordWrap,
                                    text);

        painter->setRenderHint(QPainter::TextAntialiasing);
        painter->fillRect(QRect(0, 0, width(), rect.height() + 2*border),
                          QColor(0, 0, 0, 128));
        painter->setPen(Qt::white);
        painter->drawText((width() - rect.width())/2,
                          border+metrics.descent()/4,
                          rect.width(), rect.height(),
                          Qt::AlignCenter | Qt::TextWordWrap, text);
    }
    else
        fpsCounter = 0;

    // Show legend
    const int margin = 11;
    const int padding = 6;
    text = "";

    QString tmp;

    if (showModelViewMatrix) {
        tmp.sprintf("<h4 align=\"center\">Model view matrix</h4>"
                    "<p align=\"center\"><table width=\"100\%%\">"
                    "<tr><td>%.1f</td><td>%.1f</td><td>%.1f</td><td>%.1f</td></tr>"
                    "<tr><td>%.1f</td><td>%.1f</td><td>%.1f</td><td>%.1f</td></tr>"
                    "<tr><td>%.1f</td><td>%.1f</td><td>%.1f</td><td>%.1f</td></tr>"
                    "<tr><td>%.1f</td><td>%.1f</td><td>%.1f</td><td>%.1f</td></tr>"
                    "</table>",
                    modelViewMatrix(0,0), modelViewMatrix(0,1), modelViewMatrix(0,2), modelViewMatrix(0,3),
                    modelViewMatrix(1,0), modelViewMatrix(1,1), modelViewMatrix(1,2), modelViewMatrix(1,3),
                    modelViewMatrix(2,0), modelViewMatrix(2,1), modelViewMatrix(2,2), modelViewMatrix(2,3),
                    modelViewMatrix(3,0), modelViewMatrix(3,1), modelViewMatrix(3,2), modelViewMatrix(3,3));
        text += tmp;
    }

    if (showStatistics) {
        tmp.sprintf("<h4 align=\"center\">Tracing statistics</h4>"
                    "<p align=\"center\"><table width=\"100\%%\">"
                    "<tr><td>Drawn pixels:</td><td>%.0f</td></tr>"
                    "<tr><td>Steps:</td><td>%.0f</td></tr>"
                    "<tr><td>Steps avg:</td><td>%.2f</td></tr>"
                    "<tr><td>Steps max:</td><td>%d</td></tr>"
                    "</table>",
                    drawnPixels,
                    sumOfSteps,
                    sumOfSteps / drawnPixels,
                    maximumStep);

        text += tmp;
    }

    if (text.length() > 0) {
        textDocument.setDefaultStyleSheet("* { color: #FFFFFF }");
        textDocument.setHtml(text);
        textDocument.setTextWidth(textDocument.size().width() + 20);

        rect = QRect(QPoint(0, 0), textDocument.size().toSize()
                     + QSize(2 * padding, 2 * padding));
        painter->translate(width()  - rect.width()  - margin,
                           height() - rect.height() - margin);
        painter->setPen(Qt::white);
        painter->setBrush(QColor(0, 0, 0, 128));
        painter->drawRect(rect);

        painter->translate(padding, padding);
        textDocument.drawContents(painter);
    }
}

void GLWidget::createBkgndGradient()
{
    m_bkgndGradient.setStart(QPointF(0, camera.viewportHeight()));
    m_bkgndGradient.setFinalStop(QPointF(0, camera.viewportHeight()/4));
    m_bkgndGradient.setColorAt(0, Qt::white);
    m_bkgndGradient.setColorAt(1, Qt::white);
}

void GLWidget::createOpenGLScene()
{
    makeCurrent();

    glEnable(GL_DEPTH_TEST);
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

    camera.SetScale(0.6);

    readCOEFileBin("chinese_lion_30g_15c.dat");
    bool ok = createShadersAndBOs();

    glActiveTexture(GL_TEXTURE2);
    for(int i = 0; i < 7; i++)
        if(tx_id[i] != -1) QGLWidget::deleteTexture(tx_id[i]);

    tx_id[6] = QGLWidget::bindTexture(cube_img[5],GL_TEXTURE_CUBE_MAP);

    tx_id[0] = QGLWidget::bindTexture(cube_img[0],GL_TEXTURE_CUBE_MAP_NEGATIVE_X,GL_RGB);
    tx_id[1] = QGLWidget::bindTexture(cube_img[4],GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,GL_RGB);
    tx_id[2] = QGLWidget::bindTexture(cube_img[2],GL_TEXTURE_CUBE_MAP_NEGATIVE_Z,GL_RGB);
    tx_id[3] = QGLWidget::bindTexture(cube_img[3],GL_TEXTURE_CUBE_MAP_POSITIVE_X,GL_RGB);
    tx_id[4] = QGLWidget::bindTexture(cube_img[1],GL_TEXTURE_CUBE_MAP_POSITIVE_Y,GL_RGB);
    tx_id[5] = QGLWidget::bindTexture(cube_img[5],GL_TEXTURE_CUBE_MAP_POSITIVE_Z,GL_RGB);

    mayRender = ok && (numCenters > 0);

    connect(&timer, SIGNAL(timeout()), this, SLOT(animate()));
    timer.start(0);

    fpsTime.start();
}

void GLWidget::drawBkgnd(QPainter *painter)
{
    painter->setPen(Qt::NoPen);
    painter->setBrush(m_bkgndGradient);
    painter->drawRect(rect());
}

void GLWidget::mouseMoveEvent(QMouseEvent *event)
{
    trackBall.mouseMove(event->pos());
    lightTrackBall.mouseMove(event->pos());
}

void GLWidget::mousePressEvent(QMouseEvent *event)
{
    if (event->button() & Qt::LeftButton)
        trackBall.mousePress(event->pos());
    else if (event->button() & Qt::RightButton)
        lightTrackBall.mousePress(event->pos());
}

void GLWidget::mouseReleaseEvent(QMouseEvent *event)
{
    if (event->button() == Qt::LeftButton)
        trackBall.mouseRelease(event->pos());
    else if (event->button() == Qt::RightButton)
        lightTrackBall.mouseRelease(event->pos());
}

void GLWidget::wheelEvent(QWheelEvent *event)
{
    float newScale = camera.Scale() - 0.00025 * event->delta();
    if (newScale <= 0.01)
        newScale = 0.01;
    camera.SetScale(newScale);
}

void GLWidget::keyPressEvent(QKeyEvent *event)
{
    switch(event->key())
    {
    case Qt::Key_Enter:
    case Qt::Key_Return:
        if(event->modifiers() & Qt::AltModifier)
            emit toggleFullScreen();
        break;
    case Qt::Key_Escape:
        if (isFullScreen)
            emit toggleFullScreen();
        else
            qApp->quit();
        break;
    }
}

void GLWidget::animate()
{
    update();
}

void GLWidget::switchShader(int idx)
{
    if (idx != static_cast<int>(currentShader)) {
        currentShader = idx;
        mayRender = false;
        mayRender = createShadersAndBOs();
    }
}

void GLWidget::toggleShowFPS(bool toggle)
{
    showFPS = toggle;
}

void GLWidget::toggleShowModelViewMatrix(bool toggle)
{
    showModelViewMatrix = toggle;
}

void GLWidget::toggleShowShadows(bool toggle)
{
    if (showShadows != toggle) {
        showShadows = toggle;
        mayRender = false;
        mayRender = createShadersAndBOs();
    }
}

void GLWidget::toggleShowStatistics(bool toggle)
{
    showStatistics = toggle;
}

void GLWidget::toggleAA(bool toggle)
{
    if (toggle)
        switchShader(1);
    else
        switchShader(0);
}


void GLWidget::showFileOpenDialog()
{
    timer.stop();
    QByteArray fileFormat = "dat";
    QString fileName;

    fileName = QFileDialog::getOpenFileName(this, "Open File",
                                            QDir::homePath(),
                                            QString("%1 Files (*.%2)")
                                            .arg(QString(fileFormat.toUpper()))
                                            .arg(QString(fileFormat)));
    if (!fileName.isEmpty()) {
        mayRender = false;
        readCOEFileBin(fileName);
        if (numCenters > 0)
            mayRender = true;
    }
    timer.start(0);
}

QVector3D GLWidget::calcGradient(QVector3D &p)
{
    float ac = 0;
    QVector3D g = QVector3D(0, 0, 0);

    for (int i = 0; i < numberOfCells; ++i) {
        if (cells[i]->numCenters > 0) {
            QVector3D s = QVector3D(
                        floor(i / pouDiv.w()),
                        floor((i % static_cast<int>(pouDiv.w())) / pouDiv.z()),
                        i % static_cast<int>(pouDiv.z())
                         ) * pouCellSz + minPou;
            QVector3D t = s + pouCellSz;
            QVector3D ovl = ((pouOverlap-1)*(t-s))/2;
            s -= ovl;
            t += ovl;

            if (p.x() > s.x() && p.y() > s.y() && p.z() > s.z() &&
                p.x() < t.x() && p.y() < t.y() && p.z() < t.z()) {
                QVector3D den = t-s;
                QVector3D denp2 = den*den;
                QVector3D wei = (4*(p-s)*(t-p));
                wei.setX(wei.x() / denp2.x());
                wei.setY(wei.y() / denp2.y());
                wei.setZ(wei.z() / denp2.z());
                float d = 1 - wei.x() * wei.y() * wei.z();
                float ddd = d * d * d;
                float dddd = ddd * d;
                float c = -6*dddd*d + 15*ddd*d - 10*ddd + 1;

                QVector3D gloc = QVector3D(0, 0, 0);
                for (int j = 0; j < cells[i]->numCenters; ++j)
                {
                    QVector3D df = p - cells[i]->position[j].toVector3D();
                    float r = df.length();

                    QVector3D dfs = df*df;
                    QMatrix3x3 hessian;
                    hessian(0,0) = 2*dfs.x()+dfs.y()+dfs.z(); hessian(0,1) = df.x()*df.y();             hessian(0,2) = df.x()*df.z();
                                                              hessian(1,1) = dfs.x()+2*dfs.y()+dfs.z(); hessian(1,2) = df.y()*df.z();
                                                                                                        hessian(2,2) = dfs.x()+dfs.y()+2*dfs.z();
                    hessian(1,0) = hessian(0,1);
                    hessian(2,0) = hessian(0,2);
                    hessian(2,1) = hessian(1,2);

                    hessian *= 3/r;

                    QVector4D cC = cells[i]->coefficient[j];
                    QVector3D mult;
                    mult.setX(cC.y() * hessian(0,0) + cC.z() * hessian(1,0) + cC.w() * hessian(2,0));
                    mult.setY(cC.y() * hessian(0,1) + cC.z() * hessian(1,1) + cC.w() * hessian(2,1));
                    mult.setZ(cC.y() * hessian(0,2) + cC.z() * hessian(1,2) + cC.w() * hessian(2,2));

                    gloc.setX(gloc.x() + cC.x() * 3 * r * df.x() - mult.x());
                    gloc.setY(gloc.y() + cC.x() * 3 * r * df.y() - mult.y());
                    gloc.setZ(gloc.z() + cC.x() * 3 * r * df.z() - mult.z());
                }
                gloc += QVector3D(cells[i]->polynomial.y(),
                                  cells[i]->polynomial.z(),
                                  cells[i]->polynomial.w());

                g += gloc * QVector3D(c, c, c);
                ac += c;
            }
        }
    }

    if (ac != 0)
        g /= ac;

    return g;
}

void GLWidget::estimateLipschitz()
{
    QString filename;
    filename.sprintf("lipschitz_%d_%d_%d_%d.dat",
                     static_cast<int>(pouDiv.x()),
                     static_cast<int>(pouDiv.y()),
                     static_cast<int>(pouDiv.z()), maxCentersPerCell);
    QFile file(filename);
    if (file.open(QIODevice::ReadOnly)) {
        // Load previously saved constants
        QDataStream in(&file);
        for (int i = 0; i < numberOfCells; ++i) {
            in >> cells[i]->lipschitz;
        }
        file.close();
    }
    else {
        for (int i = 0; i < numberOfCells; ++i)
            cells[i]->lipschitz = -10e10;

        for (int i = 0; i < numberOfCells; ++i) {
            float lip = -10e10;

            QVector3D s = QVector3D(
                        floor(i / pouDiv.w()),
                        floor((i % static_cast<int>(pouDiv.w())) / pouDiv.z()),
                        i % static_cast<int>(pouDiv.z())
                         ) * pouCellSz + minPou;
            QVector3D t = s + pouCellSz;
            QVector3D ovl = ((pouOverlap-1)*(t-s))/2;
            s -= ovl;
            t += ovl;

            lip = 1.3;
            if (lip > cells[i]->lipschitz )
                cells[i]->lipschitz = lip;
        }

        // Write constants to disk
        if (!file.open(QIODevice::WriteOnly)) {
            showWarning(QObject::tr("Cannot create Lipschitz file."));
            return;
        }

        // Save coefficients to disk
        QDataStream out(&file);
        for (int i = 0; i < numberOfCells; ++i)
            out << cells[i]->lipschitz;

        file.close();
    }
}

void GLWidget::readCOEFileBin(const QString &fileName)
{
    QFile file(fileName);
    if (!file.open(QIODevice::ReadOnly)) {
        showWarning(QObject::tr("Cannot open file ") + fileName);
        return;
    }

    QDataStream in(&file);
    int fileSize = file.size();
    char *p = new char[fileSize];
    in.readRawData(p, fileSize);

    maxCentersPerCell = static_cast<int>(*reinterpret_cast<float*>(p)); p += sizeof(float);

    numberOfCells = static_cast<int>(*reinterpret_cast<float*>(p)); p += sizeof(float);
    float pouDiv[3];
    for (int k = 0; k < 3; ++k) { pouDiv[k] = *reinterpret_cast<float*>(p); p += sizeof(float); }
    this->pouDiv = QVector4D(pouDiv[0], pouDiv[1], pouDiv[2], pouDiv[1] * pouDiv[2]); // 0,1,2,1*2
    pouOverlap = *reinterpret_cast<float*>(p); p += sizeof(float);

    float minPou[3];
    for (int k = 0; k < 3; ++k) { minPou[k] = *reinterpret_cast<float*>(p); p += sizeof(float); }
    this->minPou = QVector3D(minPou[0], minPou[1], minPou[2]); // 0,1,2
    float maxPou[3];
    for (int k = 0; k < 3; ++k) { maxPou[k] = *reinterpret_cast<float*>(p); p += sizeof(float); }
    this->maxPou = QVector3D(maxPou[0], maxPou[1], maxPou[2]); // 0,1,2

    pouSz = this->maxPou - this->minPou;
    pouCellSz = QVector3D(pouSz.x() / pouDiv[0], // 0,1,2
                          pouSz.y() / pouDiv[1],
                          pouSz.z() / pouDiv[2]);

    cells = new POUCell*[static_cast<int>(numberOfCells)];

    numCenters = 0;
    for (int i = 0; i < numberOfCells; ++i) {
        int curCell = static_cast<int>(*reinterpret_cast<float*>(p)) - 1;
        p += sizeof(float);
        float isValid = static_cast<int>(*reinterpret_cast<float*>(p)); p += sizeof(float);
        int curNumCen = static_cast<int>(*reinterpret_cast<float*>(p)); p += sizeof(float);
        numCenters += curNumCen;

        cells[curCell] = new POUCell(curNumCen);
        cells[curCell]->isValid = isValid;
        cells[curCell]->lipschitz = 1.0;

        for (int j = 0; j < curNumCen; ++j) {
            float centerP[3];
            for (int k = 0; k < 3; ++k) { centerP[k] = *reinterpret_cast<float*>(p); p += sizeof(float); }
            /*float centerN[3];*/
            for (int k = 0; k < 3; ++k) { /*centerN[k] = *reinterpret_cast<float*>(p);*/ p += sizeof(float); }
            float centerC[4];
            for (int k = 0; k < 4; ++k) { centerC[k] = *reinterpret_cast<float*>(p); p += sizeof(float); }

            cells[curCell]->position[j] = QVector4D(centerP[0], centerP[1], centerP[2], 0);
            cells[curCell]->coefficient[j] = QVector4D(centerC[0], centerC[1], centerC[2], centerC[3]);
        }

        float poly[4];
        for (int k = 0; k < 4; ++k) { poly[k] = *reinterpret_cast<float*>(p); p += sizeof(float); }

        cells[curCell]->polynomial = QVector4D(poly[0], poly[1], poly[2], poly[3]);
    }

    emit statusBarMessage(QObject::tr("Estimating Lipschitz constant..."));

    estimateLipschitz();

    emit statusBarMessage(QObject::tr("Loaded %1 centers").arg(numCenters));

    file.close();
}

bool GLWidget::createShadersAndBOs()
{
    makeCurrent();

    destroyShadersAndBOs();

    // Create shader program

    QString vertexShaderFile[] = {
        ":/shaders/vsphere.glsl",
        ":/shaders/vsphere.glsl",
    };
    QString fragmentShaderFile[] = {
        ":/shaders/fsphere.glsl",
        ":/shaders/fsphereaa.glsl",
    };

    vertexShader = new QOpenGLShader(QOpenGLShader::Vertex);
    if (!vertexShader->compileSourceFile(vertexShaderFile[currentShader])) {
        showWarning(tr("Unable to load vertex shader\n\n") +
                    vertexShader->log());
        return false;
    }

    QFile file(fragmentShaderFile[currentShader]);
    if (!file.open(QFile::Text | QFile::ReadOnly)) {
        showWarning(tr("Unable to open fragment shader."));
        return false;
    }
    QTextStream in(&file);
    QString str;
    str = in.readAll();
    if (showShadows) {
        str = QString("#define SHADOWS\n") + str;
    }
    file.close();

    fragmentShader = new QOpenGLShader(QOpenGLShader::Fragment);
    if (!fragmentShader->compileSourceCode(str)) {
        showWarning(fragmentShader->log());
        return false;
    }

    shaderProgram = new QOpenGLShaderProgram;
    shaderProgram->addShader(vertexShader);
    shaderProgram->addShader(fragmentShader);

    if (!shaderProgram->link()) {
        showWarning(shaderProgram->log());
        return false;
    }

    // Create buffer objects

    // Vertex buffer
    vboVertices = new QOpenGLBuffer(QOpenGLBuffer::VertexBuffer);
    if (!vboVertices->create())
        return false;
    if (!vboVertices->bind())
        return false;
    vboVertices->setUsagePattern(QOpenGLBuffer::StaticDraw);
    QVector4D vertices[4];
    vertices[0] = QVector4D(-1.0, -1.0, 0.0, 1.0);
    vertices[1] = QVector4D(-1.0,  1.0, 0.0, 1.0);
    vertices[2] = QVector4D( 1.0,  1.0, 0.0, 1.0);
    vertices[3] = QVector4D( 1.0, -1.0, 0.0, 1.0);
    vboVertices->allocate(vertices, 4 * sizeof(QVector4D));
    vboVertices->release();

    // Index buffer
    vboIndices = new QOpenGLBuffer(QOpenGLBuffer::IndexBuffer);
    if (!vboIndices->create())
        return false;
    if (!vboIndices->bind())
        return false;
    vboIndices->setUsagePattern(QOpenGLBuffer::StaticDraw);
    unsigned int indices[4] = { 0, 1, 2, 3 };
    vboIndices->allocate(indices, 4 * sizeof(unsigned int));
    vboIndices->release();

    // Texture buffer
    glGenBuffers(1, &tboP);
    glBindBuffer(GL_TEXTURE_BUFFER, tboP);
    unsigned int numSlots = maxCentersPerCell * numberOfCells + numberOfCells;
    unsigned int size = numSlots * sizeof(QVector4D); // x,y,z,numCenters
    glBufferData(GL_TEXTURE_BUFFER, size, 0, GL_STATIC_DRAW);
    QVector4D *cP = new QVector4D[numSlots];
    QVector4D *cC = new QVector4D[numSlots];
    int k = 0, l = 0;
    for (int i = 0; i < numberOfCells; ++i) {
        int curNumCen = cells[i]->numCenters;
        k = i * (maxCentersPerCell + 1);
        l = i * (maxCentersPerCell + 1);
        cP[k++] = QVector4D(qMax(1.0f, cells[i]->lipschitz), curNumCen, cells[i]->isValid, 0);
        for (int j = 0; j < curNumCen; ++j) {
            cP[k++] = cells[i]->position[j];
            cC[l++] = cells[i]->coefficient[j];
        }
        cC[l] = cells[i]->polynomial;
    }
    glBufferSubData(GL_TEXTURE_BUFFER, 0, size, cP);    

    glGenTextures(1, &texTboP);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_BUFFER, texTboP);
    glTexBuffer(GL_TEXTURE_BUFFER, GL_RGBA32F, tboP);
    glBindBuffer(GL_TEXTURE_BUFFER, 0);

    glGenBuffers(1, &tboC);
    glBindBuffer(GL_TEXTURE_BUFFER, tboC);
    size = (numSlots) * sizeof(QVector4D);
    glBufferData(GL_TEXTURE_BUFFER, size, 0, GL_STATIC_DRAW);
    glBufferSubData(GL_TEXTURE_BUFFER, 0, size, cC);

    delete []cP;
    delete []cC;

    glGenTextures(1, &texTboC);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_BUFFER, texTboC);
    glTexBuffer(GL_TEXTURE_BUFFER, GL_RGBA32F, tboC);
    glBindBuffer(GL_TEXTURE_BUFFER, 0);

    // Uniform buffer
    // Get index of uniform block
    int blockIndex = glGetUniformBlockIndex(shaderProgram->programId(),
                                            "Centers");
    if (blockIndex != -1)
    {
        // Allocate space for the buffer to contain the data for the uniform block
        GLint blockSize;
        glGetActiveUniformBlockiv(shaderProgram->programId(), blockIndex,
                                  GL_UNIFORM_BLOCK_DATA_SIZE, &blockSize);
        GLubyte *blockBuffer = new GLubyte[blockSize];
        // Query for the offset of each variable within the block
        const GLchar *names[] = { "lipCen" };
        GLuint uboIndices[1];
        glGetUniformIndices(shaderProgram->programId(), 1, names, uboIndices);
        GLint offset[1];
        glGetActiveUniformsiv(shaderProgram->programId(), 1, uboIndices,
                              GL_UNIFORM_OFFSET, offset);

        QVector4D *lc = new QVector4D[numberOfCells];
        for (int i = 0; i < numberOfCells; ++i) {
            lc[i] = QVector4D(qMax(1.0f, cells[i]->lipschitz),
                              cells[i]->numCenters,
                              cells[i]->isValid, 0);
        }

        memcpy(blockBuffer + offset[0], lc,
               qMin(numberOfCells, 4096) * sizeof(QVector4D));
        delete []lc;

        // Create the OpenGL buffer object and copy the data into it
        glGenBuffers(1, &ubo);
        glBindBuffer(GL_UNIFORM_BUFFER, ubo);
        glBufferData(GL_UNIFORM_BUFFER, blockSize, blockBuffer, GL_STATIC_DRAW);

        delete []blockBuffer;
        // Bind the buffer object to the uniform block
        glBindBufferBase(GL_UNIFORM_BUFFER, blockIndex, ubo);
    }

    return true;
}

void GLWidget::destroyShadersAndBOs()
{       
    makeCurrent();

    delete vertexShader;
    vertexShader = NULL;

    delete fragmentShader;
    fragmentShader = NULL;

    if (shaderProgram) {
        shaderProgram->release();
        delete shaderProgram;
        shaderProgram = NULL;
    }

    if (ubo != 0) {
        glDeleteBuffers(1, &ubo);
        ubo = 0;
    }

    if (tboP != 0) {
        glBindBuffer(1, tboP);
        glDeleteBuffers(1, &tboP);
        glDeleteTextures(1, &texTboP);
        tboP = 0;
        texTboP = 0;
    }

    if (tboC != 0) {
        glBindBuffer(1, tboC);
        glDeleteBuffers(1, &tboC);
        glDeleteTextures(1, &texTboC);
        tboC = 0;
        texTboC = 0;
    }

    if (vboVertices) {
        vboVertices->release();
        delete vboVertices;
        vboVertices = NULL;
    }

    if (vboIndices) {
        vboIndices->release();
        delete vboIndices;
        vboIndices = NULL;
    }
}
