/*
Example code for the technique presented in the paper

GPU-based Sphere Tracing for Radial Basis Functions Implicits
(Batagelo, H. C.; Gois, J. P.). International Journal of Image and Graphics,
14(1-2), 2014, DOI 10.1142/S0219467814500041

Copyright (c) 2014, Harlen Costa Batagelo
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the copyright holder may not be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL HARLEN COSTA BATAGELO BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef GLWIDGET_H
#define GLWIDGET_H

#include <QGLWidget>
#include <QKeyEvent>
#include <QDialog>
#include <QTimer>
#include <QtOpenGL>
#include <QOpenGLFunctions_3_3_Compatibility>

#include <iostream>
#include <fstream>
#include <limits>

#include "camera.h"
#include "light.h"
#include "material.h"
#include "trackball.h"

class POUCell
{
public:
    POUCell(int centers)
    {
        numCenters = centers;
        position = new QVector4D[numCenters];
        coefficient = new QVector4D[numCenters];
        lipschitz = 1.0;
    };
    virtual ~POUCell()
    {
        delete[] position;
        delete[] coefficient;
    }

    int numCenters; // Number of centers
    QVector4D *position; // Array of center positions (1st w is # of centers)
    QVector4D *coefficient; // Array of center coefficients
    QVector4D polynomial; // Polynomial coefficients
    float lipschitz; // Lipschitz constant at this cell
    bool isValid; // False if it contains a dummy center only
};

class GLWidget : public QGLWidget, protected QOpenGLFunctions_3_3_Compatibility
{
    Q_OBJECT
public:
    explicit GLWidget(QWidget *parent = 0);
    virtual ~GLWidget();

    bool isFullScreen;

signals:
    void toggleFullScreen();
    void statusBarMessage(QString ns);
    
public slots:
    void showFileOpenDialog();
    void animate();
    void switchShader(int idx);
    void toggleShowFPS(bool toggle);
    void toggleShowModelViewMatrix(bool toggle);
    void toggleShowShadows(bool toggle);
    void toggleShowStatistics(bool toggle);
    void toggleAA(bool toggle);
    
protected:
    void initializeGL();
    void resizeGL(int width, int height);
    void paintEvent(QPaintEvent *event);
    void mouseMoveEvent(QMouseEvent *event);
    void mousePressEvent(QMouseEvent *event);
    void mouseReleaseEvent(QMouseEvent *event);
    void wheelEvent(QWheelEvent *event);
    void keyPressEvent(QKeyEvent *event);

private:
    void createBkgndGradient();
    void createOpenGLScene();
    void drawBkgnd(QPainter *painter);
    void drawOpenGLScene();
    void displayInfo(QPainter *painter);

    void estimateLipschitz();
    QVector3D calcGradient(QVector3D &p);
    void readCOEFileBin(const QString &fileName);
    void readCOEFile(const QString &fileName);
    bool createShadersAndBOs();
    void destroyShadersAndBOs();

    bool mayRender;

    GLuint tboP, texTboP;
    GLuint tboC, texTboC;

    GLuint ubo;
    QOpenGLBuffer *vboVertices;
    QOpenGLBuffer *vboIndices;
    GLuint fbo, rbo, texo[2];

    QOpenGLShader *vertexShader;
    QOpenGLShader *fragmentShader;
    QOpenGLShaderProgram *shaderProgram;
    unsigned int currentShader;

    int texID[2];

    QMatrix4x4 modelViewMatrix;
    QMatrix4x4 projectionMatrix;

    QVector2D minSphere;

    Camera camera;
    Light light;
    Material material;

    TrackBall trackBall;
    TrackBall lightTrackBall;

    QTimer timer;

    QTime fpsTime;
    double fps;
    int fpsCounter;

    QLinearGradient m_bkgndGradient;

    QVector3D lastPosition;

    bool showFPS;
    bool showModelViewMatrix;
    bool showShadows;
    bool showStatistics;

    // Statistics
    double drawnPixels;
    double sumOfSteps;
    int maximumStep;

    int maxCentersPerCell; // Maximum number of centers per cell
    int numberOfCells; // Number of cells
    QVector4D pouDiv; // Division of cells along each axis
    float pouOverlap; // Cell overlap factor
    QVector3D minPou, maxPou; // Centers bounding box
    QVector3D pouSz; // Bounding box size (maxPou - minPou)
    QVector3D pouCellSz; // Size of each POU cell

    POUCell **cells; // Array of partition-of-unity cells

    int numCenters;
    QVector4D *centersP;
    QVector3D *centersN;
    QVector4D *centersC;

    QImage cube_img[6];
    int tx_id[7];
};

#endif // GLWIDGET_H
